# -*- coding: utf-8 -*-
"""Definition of State class used in pyrodeo.

The State class holds density, velocity and sound speed for a pyrodeo simulation
"""

from __future__ import print_function

import numpy as np

class State(object):
    """Construct state holding density, velocity and sound speed for a pyrodeo simulation.

    Args:
        dens (ndarray): 2D ndarray containing density.
        velx (ndarray): 2D ndarray containing x velocity.
        vely (ndarray): 2D ndarray containing y velocity.
        soundspeed (ndarray): 2D ndarray containing sound speed.

    Note:
        No checks are performed whether density, velocity and sound speed are valid arrays. They should all have the same shape, the same as the arrays of :class:`.Coordinates`.

    The following public attributes are available:

    Attributes:
        dens (ndarray): 2D ndarray containing density.
        velx (ndarray): 2D ndarray containing x velocity.
        vely (ndarray): 2D ndarray containing y velocity.
        soundspeed (ndarray): 2D ndarray containing sound speed.
        no_ghost (ndarray): 2D ndarray flagging whether a cell is a ghost cell (=0) or an internal cell (=1)

    """

    def __init__(self, dens, velx, vely, soundspeed):
        self.dens = dens
        self.velx = velx
        self.vely = vely
        self.soundspeed = soundspeed

        self.no_ghost = np.full(np.shape(dens), 1)
        self.no_ghost[:2,:]  = 0
        self.no_ghost[-2:,:] = 0
        if len(dens[0,:]) > 1:
            self.no_ghost[:,:2]  = 0
            self.no_ghost[:,-2:] = 0

    @classmethod
    def from_dims(cls, dims):
        """Construct State from grid dimensions.

        Construct State given grid dimensions, creating arrays of the correct size with standard (physical) values.

        Args:
            dims (int, int): Dimensions of the grid in x and y

        """
        if len(dims) != 2:
            raise TypeError('Expexted dimensions to have two elements')
        if (dims[0] < 1 or dims[1] < 1):
            raise ValueError('Need both dimensions to be larger than zero')

        dens = np.full(dims, 1.0)
        velx = np.full(dims, 0.0)
        vely = np.full(dims, 0.0)
        soundspeed = np.full(dims, 1.0)

        return cls(dens, velx, vely, soundspeed)

    @classmethod
    def copy(cls, other_state):
        """Construct state from other State.

        Set this instance of State equal to an other State, performing an explicit copy.

        Args:
            other_state (State): State from which to copy.

        """
        dens = np.copy(other_state.dens)
        velx = np.copy(other_state.velx)
        vely = np.copy(other_state.vely)
        soundspeed = np.copy(other_state.soundspeed)

        return cls(dens, velx, vely, soundspeed)

    def transpose(self):
        """Transpose all fields."""
        self.dens = np.transpose(self.dens)
        self.velx = np.transpose(self.velx)
        self.vely = np.transpose(self.vely)
        self.soundspeed = np.transpose(self.soundspeed)
        self.no_ghost = np.transpose(self.no_ghost)

    def swap_velocities(self):
        """Swap x and y velocities."""
        tmp = self.velx
        self.velx = self.vely
        self.vely = tmp



# -*- coding: utf-8 -*-
"""Defines the Simulation class holding a pyrodeo simulation
"""

from __future__ import print_function

import numpy as np
import h5py
from pyrodeo.state import State
from pyrodeo.coords import Coordinates
from pyrodeo.param import Param
from pyrodeo.hydro import Hydro

class Simulation(object):
    """Construct pyrodeo simulation.

    Construct simulation from existing instances of :class:`.Param`, :class:`.Coordinates` and :class:`.State` and given simulation time.

    Args:
        param (:class:`.Param`): Valid Param object, containing simulation parameters.
        coords (:class:`.Coordinates`): Valid Coordinates object, containing x and y coordinates
        state (:class:`.State`): Valid State object, containing current density, velocity and sound speed.
        t (float): Simulation time.
        direc (:obj:`str`, optional): Output directory.

    Note:
        While it is possible to set up a simulation using the basic constructor, no checks are performed whether the instances of :class:`.Param`, :class:`.Coordinates` and :class:`.State` are valid. It is safer to use :meth:`.from_geom` which will set up the necessary valid instances.

    The following attributes and methods are available:

    Attributes:
        state (:class:`.State`): State object holding density, velocity and sound speed.
        coords (:class:`.Coordinates`): Coordinates object holding x and y coordinates.
        param (:class:`.Param`): Param object holding simulation parameters.
        t (float): Current simulation time.
        direc (string): Output directory.

    """

    def __init__(self, param, coords, state, t, direc='./'):
        self.state = State(state.dens, state.velx, state.vely, state.soundspeed)
        self.coords = Coordinates(coords.x, coords.y)
        self.param = Param(param)
        self.t = t
        self.direc = direc

    @classmethod
    def from_geom(cls,
                  geometry,
                  dimensions=(100, 1),
                  domain=([-0.5, 0.5], []),
                  direc='./'):
        """Construct simulation from geometry.

        Construct simulation class from geometry, grid dimensions and domain size. All other parameters will be set to defaults. Simulation time is set to zero.

        Args:
            geometry (string): 'cart' (Cartesian coordinates), 'sheet' (shearing sheet), or 'cyl' (cylindrical coordinates).
            dimensions (:obj:`(int,int)`,optional): Grid dimensions in x and y.
            domain (:obj:`[(float,float),(float,float)]`,optional): Domain boundaries in x and y.
            direc (:obj:`str`,optional): Output directory, defaults to current directory.

        """
        dt, param = Param.from_geom(geometry).to_list()

        coords = Coordinates.from_dims(dimensions, domain)

        state = State.from_dims((len(coords.x[:,0]),
                                 len(coords.x[0,:])))
        t = 0.0

        return cls(param[0], coords, state, t, direc=direc)

    @classmethod
    def from_checkpoint(cls, direc, n=None):
        """Constructor for simulation class from checkpoint.

        Construct simulation class from previously saved checkpoint.

        Args:
            direc (str): Path to 'rodeo.h5'. This will be the output directory as well.
            n (:obj:`int`,optional): Index of checkpoint to restore. If left None, restore last saved checkpoint.
        """
        with h5py.File(direc + '/rodeo.h5', "r") as hf:
            # Restart from last checkpoint
            if n is None:
                g = hf.keys()
                n = len(g) - 3

            gc = hf.get("coords")
            x = np.array(gc.get('x'))
            y = np.array(gc.get('y'))

            param = hf.get("param")

            g = hf.get("checkpoint{}".format(n))
            dens = np.array(g.get('dens'))
            velx = np.array(g.get('velx'))
            vely = np.array(g.get('vely'))
            soundspeed = np.array(g.get('soundspeed'))

            coords = Coordinates(x, y)
            state = State(dens, velx, vely, soundspeed)

            t = g.attrs['time']
            print("Restoring from checkpoint at t = {}".format(t))

            return cls(param[0], coords, state, t, direc)

    def __str__(self):
        """Information on simulation."""
        return 'Simulation class, part of pyrodeo.\n \
    Current time: {}\n'.format(self.t) + str(self.coords) + str(self.param)

    def evolve(self, checkpoints,
               source_func = None, source_param = None,
               new_file=False):
        """Evolve simulation over a list of checkpoints.

        Args:
            checkpoints (ndarray): List of times to save checkpoints.
            Source (:obj:`callable`, optional): Extra source terms. Must be of the form f(t, coords, state, source_param) and return three source terms of the same shape as state.dens: density source term, x velocity source term and y velocity source term.
            source_param (:obj:`ndarray`, optional): Parameters for extra source term. Will be passed to the Source function.
            new_file (:obj:`bool`,optional): If true, create new output file 'rodeo.h5', otherwise append to file if it exists.

        """
        solver = Hydro(self.param, self.coords)
        if new_file is True:
            self.checkpoint(new_file=True)
        for tmax in checkpoints:
            if self.t < tmax:
                self.t, self.state = \
                    solver.evolve(self.t,
                                  tmax,
                                  self.coords,
                                  self.param,
                                  self.state,
                                  source_func,
                                  source_param)
                self.checkpoint()
            else:
                print("Skipping checkpoint t = {}".format(tmax))

    def checkpoint(self, new_file=False):
        """Save checkpoint in 'rodeo.h5'

        Args:
            new_file (:obj:`bool`,optional): If true, create new output file 'rodeo.h5', otherwise append to file if it exists.

        """
        mode = "a"
        if new_file is True:
            mode = "w"

        print("Checkpoint at t = {}".format(self.t))

        with h5py.File(self.direc + '/rodeo.h5', mode) as hf:
            n = 0

            # Write coordinates and parameters
            if new_file is True:
                gc = hf.create_group("coords")
                gc.create_dataset("x", data=self.coords.x)
                gc.create_dataset("y", data=self.coords.y)

                param_dtype, param = self.param.to_list()
                param_array = np.array(param, dtype=param_dtype)
                param_dset = \
                  hf.create_dataset("param", (len(param),), dtype=param_dtype)
                param_dset[...] = param_array
            else:
                # Number of previous saves + 1
                g = hf.keys()
                n = len(g) - 2

            # Write current data
            g1 = hf.create_group("checkpoint{}".format(n))
            g1.attrs['time'] = self.t
            g1.create_dataset("dens", data=self.state.dens)
            g1.create_dataset('velx', data=self.state.velx)
            g1.create_dataset('vely', data=self.state.vely)
            g1.create_dataset('soundspeed', data=self.state.soundspeed)



# -*- coding: utf-8 -*-
"""Simulation parameters used in pyrodeo.

The Param class holds parameters necessary for a pyrodeo simulation.
"""


from __future__ import print_function

import numpy as np
import h5py

class Param(object):
    """Create instance from list of parameters.

    Args:
        param_list ([str, float, float, float, str, str]): List of parameters; geometry (string), courant (float), fluxlimiter (float), Omega (float), boundaries x (string) and boundaries y (string)

    Note:
        The validity of the parameters is not checked.

    The following public attributes are available:

    Attributes:
        geometry (string): 'cart' (Cartesian coordinates), 'sheet' (shearing sheet) or 'cyl' (cylindrical coordinates)
        courant (float): Courant number, should be > 0 and < 1.
        limiter_param (float): Limiter parameter. Should be between 1 (minmod) and 2 (superbee).
        min_dens (float): Minimum density to switch to HLL solver to remain positive.
        Omega (float): Frame rotation rate. Ignored in Cartesian coordinates, should be unity in a shearing sheet calculation and corresponds to the angular velocity of the coordinate frame in cylindrical coordinates.
        boundaries (string, string): Boundary conditions in x and y: 'nonreflect', 'reflect', or 'periodic'.

    """

    def __init__(self, param_list):
        self.geometry      = param_list[0]
        self.courant       = param_list[1]
        self.limiter_param = param_list[2]
        self.min_dens      = param_list[3]
        self.Omega         = param_list[4]
        self.boundaries    = [param_list[5], param_list[6]]

    @classmethod
    def from_geom(cls,
                  geometry,
                  boundaries=['reflect', 'reflect']):
        """Initialization from geometry and boundary conditions.

        Construct Parameter object from geometry and boundary conditions. All other parameters are set to standard values. Check if geometry and boundary conditions are valid.

        Args:
            geometry (string): 'cart' (Cartesian coordinates), 'sheet' (shearing sheet) or 'cyl' (cylindrical coordinates).
            boundaries (string, string): boundary conditions; 'reflect', 'nonreflect' or 'periodic'

        """
        courant = 0.4
        limiter_param = 1.0
        min_dens = 1.0e-6
        Omega = 1.0
        if (geometry != 'cart' and
            geometry != 'sheet' and
            geometry != 'cyl'):
            raise ValueError('Invalid geometry')
        if len(boundaries) != 2:
            raise TypeError('Expected boundaries to have two elements')
        if (boundaries[0] != 'reflect' and
            boundaries[0] != 'nonreflect' and
            boundaries[0] != 'periodic'):
            raise ValueError('Invalid x boundary')
        if (boundaries[1] != 'reflect' and
            boundaries[1] != 'nonreflect' and
            boundaries[1] != 'periodic'):
            raise ValueError('Invalid y boundary')

        return cls([geometry,
                    courant,
                    limiter_param,
                    min_dens,
                    Omega,
                    boundaries[0],
                    boundaries[1]])

    def to_list(self):
        """Convert parameters to list.

        Return list of parameters together with list of types. Used for HDF5 output.

        Returns:
            : array of types, list of parameters.

        """
        param_dtype = np.dtype([('geom', h5py.special_dtype(vlen=str)),
                                ('courant', np.float),
                                ('limiter_param', np.float),
                                ('min_dens', np.float),
                                ('Omega', np.float),
                                ('boundary_x', h5py.special_dtype(vlen=str)),
                                ('boundary_y', h5py.special_dtype(vlen=str))])

        return param_dtype, [(self.geometry,
                              self.courant,
                              self.limiter_param,
                              self.min_dens,
                              self.Omega,
                              self.boundaries[0],
                              self.boundaries[1])]



# -*- coding: utf-8 -*-
"""Defines the Hydro class performing hydrodynamic updates of the state.
"""

from __future__ import print_function

import numpy as np
from pyrodeo.state import State
from pyrodeo.linear_advection import LinearAdvection
from pyrodeo.roe import Roe

class Hydro:
    """Construct class for hydrodynamic updates.

    Construct from existing instances of :class:`.Param` and :class:`.Coordinates`. It constructs a :class:`.LinearAdvection` instance which will deal with orbital advection, and a :class:`.Roe` instance dealing with hydrodynamics of residual velocities.

    Args:
        param (:class:`.Param`): Valid Param object, containing simulation parameters.
        coords (:class:`.Coordinates`): Valid Coordinates object, containing x and y coordinates. Used to calculate orbital advection velocity.

    The following attributes and methods are available:

    Attributes:
        orbital_advection (:class:`.LinearAdvection`): Instance of :class:`.LinearAdvection` class used to do orbital advection.
        roe (:class:`.Roe`): Instance of :class:`.Roe` class dealing with hydrodynamics of residual velocities.

    """
    def __init__(self, param, coords):
        v_adv = 0.0*coords.x
        if param.geometry == 'sheet':
            v_adv = (-1.5*coords.x).transpose()
        if param.geometry == 'cyl':
            v_adv = (np.power(coords.x, -1.5) - param.Omega).transpose()
        self.orbital_advection = LinearAdvection(v_adv, param.limiter_param)
        self.roe = Roe(param.limiter_param, param.min_dens)

    def calc_time_step(self, geometry, coords, state):
        """Calculate time step obeying the CFL condition.

        Args:
            geometry (str): 'cart', 'sheet' or 'cyl'.
            coords (:class:`.Coordinates`): Valid :class:`.Coordinates` object, containing x and y coordinates.
            state (:class:`.State`): Valid :class:`.State` object, containing density and velocity.

        Returns:
            float: Maximum time step obeying the CFL condition.

        """
        dtx = coords.dxy[0]/np.max(np.abs(state.velx) + state.soundspeed)
        dty = dtx
        if len(state.dens[0,:]) > 1:
            cs = state.soundspeed
            if geometry == 'cyl':
                cs = state.soundspeed/coords.x
            dty = coords.dxy[1]/np.max(np.abs(state.vely) + cs)

        return np.min([dtx, dty])

    def shear_periodic_boundaries(self, t, coords, state):
        """Set shear-periodic boundary conditions.

        In the shearing sheet geometry, the x direction can be quasi-periodic, i.e. periodic but modified for the shear. Imagine neighbouring sheets shearing past the center sheet.

        Args:
            t (float): Current simulation time. Used to calculate over what distance neighbouring sheets have shifted since t = 0.
            coords (:class:`.Coordinates`): Valid :class:`.Coordinates` object, containing x and y coordinates.
            state (:class:`.State`): Valid :class:`.State` object, containing density and velocity.
        """
        nx = np.shape(state.dens)[0]

        # Create State made just of ghost zones
        dens = np.concatenate((state.dens[nx-4:nx-2,:], state.dens[2:4,:]))
        velx = np.concatenate((state.velx[nx-4:nx-2,:], state.velx[2:4,:]))
        vely = np.concatenate((state.vely[nx-4:nx-2,:], state.vely[2:4,:]))
        temp_state = State(dens, velx, vely, vely)

        # Advection velocity is speed of next box
        v_adv = 0.0*dens + 1.5*(coords.x[nx-2,0] + coords.x[nx-3,0])
        v_adv[2:4,:] = -v_adv[2:4,:]

        # Linear advection since t = 0 when solution was periodic
        temp_state.transpose()
        v_adv = v_adv.transpose()

        la = LinearAdvection(v_adv, self.orbital_advection.sb)
        la.step(t, coords.dxy[1], temp_state)

        temp_state.transpose()

        # Store in ghost zones
        state.dens[:2,:] = temp_state.dens[:2,:]
        state.velx[:2,:] = temp_state.velx[:2,:]
        state.vely[:2,:] = temp_state.vely[:2,:]
        state.dens[nx-2:,:] = temp_state.dens[2:,:]
        state.velx[nx-2:,:] = temp_state.velx[2:,:]
        state.vely[nx-2:,:] = temp_state.vely[2:,:]

    def preprocess(self, coords, param, state, direction):
        """Modify state to quasi-cartesian form and calculate geometric source terms.

        Isothermal hydrodynamics allows for a generic form of the equations in all geometries, subject only to modifications in the geometric source terms.

        Args:
            coords (:class:`.Coordinates`): Valid :class:`.Coordinates` object, containing x and y coordinates.
            param (:class:`.Param`): Valid :class:`.Param` object, containing simulation parameters.
            state (:class:`.State`): Valid :class:`.State` object, containing density and velocity.
            direction (int): 0 (integrating x) or 1 (integrating y).

        Returns:
            ndarray: Geometric source term of the same shape as state.dens.

        """
        source = 0.0*state.dens

        if param.geometry == 'sheet':
            if direction == 0:
                state.vely += 0.5*param.Omega*coords.x
                source = 2.0*state.dens*param.Omega*(state.vely -
                                                     0.5*param.Omega*coords.x)
        if param.geometry == 'cyl':
            if direction == 0:
                state.dens *= coords.x
                state.vely = coords.x*coords.x*(state.vely +
                                                np.power(coords.x, -1.5))
                source = \
                    (state.vely*state.vely/coords.x -
                     1.0)*state.dens/(coords.x*coords.x) + \
                state.soundspeed*state.soundspeed*state.dens/coords.x
            if direction == 1:
                state.soundspeed /= coords.x

        if direction == 1:
            source = source.transpose()
            state.transpose()
            state.swap_velocities()

        return source

    def postprocess(self, coords, param, state, direction):
        """Inverse of :meth:`.preprocess`.

        Reverse modifications by :meth:`.preprocess`.

        Args:
            coords (:class:`.Coordinates`): Valid :class:`.Coordinates` object, containing x and y coordinates.
            param (:class:`.Param`): Valid :class:`.Param` object, containing simulation parameters.
            state (:class:`.State`): Valid :class:`.State` object, containing density and velocity.
            direction (int): 0 (integrating x) or 1 (integrating y).

        """
        if direction == 1:
            state.transpose()
            state.swap_velocities()

        if param.geometry == 'sheet':
            if direction == 0:
                state.vely -= 0.5*param.Omega*coords.x

        if param.geometry == 'cyl':
            if direction == 0:
                state.dens /= coords.x
                state.vely = state.vely/(coords.x*coords.x) - \
                np.power(coords.x, -1.5)
            if direction == 1:
                state.soundspeed *= coords.x

        if np.min(state.dens) < 0.0:
            i = np.unravel_index(np.argmin(state.dens, axis=None),
                                 state.dens.shape)
            print("Negative density encountered after integrating" +
                  " direction {} at x = {}, y = {}".format(direction,
                                                           coords.x[i],
                                                           coords.y[i]))

    def evolve(self, t, t_max, coords, param, state,
               source_func, source_param):
        """Evolve state from t to tmax.

        Args:
            t (float): Current simulation time.
            t_max (float): Simulation time to reach before stopping.
            coords (:class:`.Coordinates`): Valid :class:`.Coordinates` object, containing x and y coordinates.
            param (:class:`.Param`): Valid :class:`.Param` object, containing simulation parameters.
            state (:class:`.State`): Valid :class:`.State` object, containing density and velocity.
            source_func (callable): Function returning any extra source terms (non-geometric). It should accept the following arguments: t, coords, state, source_param and return density source, x velocity source, y velocity source.
            source_param (array-like): Extra parameters for source_func.

        Returns:
            (tuple): Tuple consisting of:

                t (float): new simulation time (= t_max if no problems encountered).

                :class:`.State`: Updated :class:`.State`.

        """
        # Evolve until t = t_max
        while (t < t_max):
            # Calculate time step
            dt = param.courant*self.calc_time_step(param.geometry,
                                                   coords, state)
            # End exactly on t_max
            if (t + dt > t_max):
                dt = t_max - t

            # Copy state in case something goes wrong
            old_state = State.copy(state)

            # Continue until everything OK
            success = False
            while success is False:
                print("Current time: {}, time step: {}".format(t, dt))

                # Shear periodic x boundaries if necessary
                if (param.geometry == 'sheet' and
                    param.boundaries[0] == 'shear periodic'):
                    self.shear_periodic_boundaries(t, coords, state)

                # Dimensional split: do all dimensions independently
                for dim in (0,1):
                    if np.shape(state.dens)[dim] > 1:
                        source = self.preprocess(coords, param, state, dim)

                        # Hydrodynamic update
                        self.roe.step(dt, coords.dxy[dim], state,
                                      source, param.boundaries[dim])

                        # Orbital advection
                        if dim == 1:
                            self.orbital_advection.step(dt,
                                                        coords.dxy[dim],
                                                        state)

                        self.postprocess(coords, param, state, dim)

                # Check if density positive
                success = True
                mindens = np.min(state.dens)

                # If not, restore state and try with smaller time step
                if mindens < 0.0:
                    state = State.copy(old_state)
                    dt = 0.5*dt
                    success = False

            # Integrate extra source terms (ignoring ghost zones!)
            if source_func is not None:
                sdens, svelx, svely = \
                  source_func(t, coords, state, source_param)
                state.dens += dt*sdens*state.no_ghost
                state.velx += dt*svelx*state.no_ghost
                state.vely += dt*svely*state.no_ghost

            # Update simulation time
            t = t + dt

        return t, state
